#include "textlcd.h"
#include "circbuff.h"
#include "delay.h"	/* prototype for delay() */
#include <string.h>
#include <stdio.h>
#include <stdbool.h>

//#define Z80_KIO 0 //--> Makefile

#if Z80_KIO
  #include "z80_kio.h"
#else
  #include "z84c015.h"
#endif

volatile unsigned char counter;
volatile unsigned long seconds;

circbuff_t cb_sioa_rx, cb_sioa_tx;

/*
 * Interrupt Vector Table
 */
typedef void (*intvec_tbl_t)(void);

#define INT_VEC_CTC 0x00
#define INT_VEC_SIO 0x10

extern intvec_tbl_t int_vector_table[32];	// Platz für 32 Vektoren (64 byte)

/*
 * Hier kann ein Breakpoint eingefügt werden
 */
void breakpoint_func(unsigned char param_in_a)
{
	(void) param_in_a;
	__asm__ ( "nop" );
}


/*
	9,8304 MHz
	1 Taktzyklus = 0,1017 us
	9830,4 Taktzyklen
	26 Taktzyklen mit Sprung
	379 Durchläufe
*/
#define LOOPCOUNT ((F_CPU / 1000 / 24) - 30)

void loop_ms(unsigned int cnt) {
	(void) cnt;
	__asm__ (
		"delay_loop:             \n"
		"	dec hl            ;[ 6] \n"
		"	ld a, h           ;[ 4] \n"
		"	or l              ;[ 4] \n"
		"	jp nz, delay_loop ;[10]   "
		                         //24*cnt + 10 for ret
	);
}

void delay1ms(void) {
	loop_ms(LOOPCOUNT);	// ld hl, jp: 20 cycles
}

void delay(unsigned long ms) {
	for (unsigned long i = 0; i < ms; i++)
		delay1ms();
}

void NMI_InterruptHandler(void) __critical __interrupt {
}

// CTC Ch1 ISR
void ctc_ch1_isr(void) __critical __interrupt(0)
{
	unsigned char i = counter + 1;

	if (i == 200) {
		seconds += 1;
		i = 0;
	}
	counter = i;
}

/*--------------------------------------------------------------------------*/

bool stat_tx_running;

// SIO A Transmit Buffer Empty ISR
void sioa_tx_isr(void) __critical __interrupt(1)
{
	unsigned char value;

	if (circbuff_read(&cb_sioa_tx, &value)) {
		SIO_A_DATA = value;
	} else {
		SIO_A_CMD = 0x28; // 0b00101000 Reset Tx Pending
		stat_tx_running = false;
	}
}

// SIO A Receive Character Available ISR
void sioa_rx_isr(void) __interrupt {
    circbuff_write(&cb_sioa_rx, SIO_A_DATA);
}

// SIO A External/Status Change
void sioa_ext_isr(void) __interrupt {
	SIO_A_CMD = 0x10; // 0b00010000 Reset External/Status Interrupts
}

// SIO A Special Receive Condition
void sioa_srx_isr(void) __interrupt {
	SIO_A_CMD = 0x30; // 0b00110000 Error Reset
}

int sio_getc(void)
{
	unsigned char value;

	if (circbuff_read(&cb_sioa_rx, &value)) {
		return value;
	} else {
		return EOF;
	}
}

void sio_putc(unsigned char value)
{
	__asm__ ( "di" );
	if (stat_tx_running) {
		while (!circbuff_write(&cb_sioa_tx, value))
			__asm__ ( "ei	\n"
				  "nop	\n"
				  "di	  ");
	} else {
		SIO_A_DATA = value;
		stat_tx_running = true;
	}
	__asm__ ( "ei" );
}

/*--------------------------------------------------------------------------*/

#if 0 // polling functions, not used
void waitTx(void) {
	unsigned char status;
	do {
		status = SIO_A_CMD;
	} while (!(status & 0x04));
}

void waitRx(void) {
	unsigned char status;
	do {
		status = SIO_A_CMD;
	} while (!(status & 0x01));
}

unsigned char readChar(void) {
	waitRx();
	return SIO_A_DATA;
}

void sendChar(unsigned int ch) {
	waitTx();
	SIO_A_DATA = ch;
}

void wait(void) {
	delay(1000);
}
#endif

/*--------------------------------------------------------------------------*/

void init_i_reg(unsigned char vh)
{
	(void) vh; // argument is in register a

	__asm__ (
		"ld i,a      \n"
		"im 2        \n"
	);
}

void init_int_vectors(void)
{
	int_vector_table[INT_VEC_CTC+1] = ctc_ch1_isr;
	int_vector_table[INT_VEC_SIO+4] = sioa_tx_isr;
	int_vector_table[INT_VEC_SIO+5] = sioa_ext_isr;
	int_vector_table[INT_VEC_SIO+6] = sioa_rx_isr;
	int_vector_table[INT_VEC_SIO+7] = sioa_srx_isr;

	init_i_reg(((unsigned) int_vector_table) >> 8);
}

#if Z80_KIO
void init_kio(void)
{
//	KIO_CMD = 0x0B;

	// CTC CH0 Baudrate 19200
	unsigned char ch0_time_constant = 2;        //9,8304E6 / 16 / 16 / 19200
	CTC_CH0 = /* 0x05; */ 0b00000101 ;             //Prescaler Value = 16
	CTC_CH0 = ch0_time_constant;

	// CTC CH1 Systick 5 ms
	unsigned char ch1_time_constant = 192;
	//CTC_CH1 = 0x25; // 0b0010 0101
	CTC_CH1 = 0xA7; // 0b1010 0101 Interrupt enabled
	CTC_CH1 = ch1_time_constant;

	// CTC Interrupt Vektor muss in CH0 geschrieben werden
	CTC_CH0 = 0x00;

	// SIO A mit Interrupts
	SIO_B_CMD = 0x02; // WR2
	SIO_B_CMD = 0x20; // Interrupt Vektor
	SIO_B_CMD = 0x01; // WR1
	SIO_B_CMD = 0x04; // 0b00000100 Status affects Vector
	// SIO A Standard Init
	SIO_A_CMD = 0x18; // 0b00011000 Channel Reset
	SIO_A_CMD = 0x04; // WR4
	SIO_A_CMD = 0x44; // 0x01000100 16x Clock, 1 Stopbit, keine Parität
	SIO_A_CMD = 0x05; // WR5
	SIO_A_CMD = 0x68; // 0b01101000 Tx 8 Datenbits, Tx Enable
	SIO_A_CMD = 0x03; // WR3
	SIO_A_CMD = 0xC1; // 0b11000001 Rx 8 Datenbits, Rx Enable
	// SIO A Interrupt Enable
	SIO_A_CMD = 0x01; // WR1
	SIO_A_CMD = 0x1B; // 0b00011011 Interrupt Enable

	// PIO A
	PIO_A_CMD = 0xCF; // 0b11001111 Mode 3
	PIO_A_CMD = 0x00; // Output

	// PIA (Port C), All Outputs
	PIA_C_CMD = 0x00;

}
#else
void init_z84c015(void)
{
/*
 *	TC_BAUD = CLK / SIO_CLK / BAUD
 *	TC_BAUD = 3072000 / 16 / 19200 = 10
 * 	TC_SYSTICK = F_CPU / CTC_PRESCALER / 1000 * 5
 * 	TC_SYSTICK = F_CPU / 256 / 1000 * 5 = 192        (F_CPU = 9.8304E6)
 */

	// CTC CH1 Systick 5 ms
	//CTC_CH1 = 0x25; // 0b0010 0101
	CTC_CH1 = 0xA7;   // 0b1010 0101  Int enabled, Timer Mode, /256, TC follows
	CTC_CH1 = F_CPU / 256 / 1000 * 5;

	// CTC CH2 Baudrate 19200
	CTC_CH2 = 0b01000111 ;	         //Counter Mode, TC follows, Reset
	CTC_CH2 = 3072000 / 16 / BAUD;

	// CTC Interrupt Vektor muss in CH0 geschrieben werden
	CTC_CH0 = INT_VEC_CTC<<1;

	// SIO A mit Interrupts
	SIO_B_CMD = 0x01; // WR1
	SIO_B_CMD = 0x04; // 0b00000100 Status affects Vector
	SIO_B_CMD = 0x02; // WR2
	SIO_B_CMD = INT_VEC_SIO<<1; // Interrupt Vektor
	// SIO A Standard Init
	SIO_A_CMD = 0x18; // 0b00011000 Channel Reset
	SIO_A_CMD = 0x04; // WR4
	SIO_A_CMD = 0x44; // 0x01000100 16x Clock, 1 Stopbit, keine Parität
	SIO_A_CMD = 0x05; // WR5
	SIO_A_CMD = 0x68; // 0b01101000 Tx 8 Datenbits, Tx Enable
	SIO_A_CMD = 0x03; // WR3
	SIO_A_CMD = 0xC1; // 0b11000001 Rx 8 Datenbits, Rx Enable
	// SIO A Interrupt Enable
	SIO_A_CMD = 0x01; // WR1
	SIO_A_CMD = 0x1A; // 0b00011010 Rx, Tx Int Enable, Ext/Stat Int Disable

	// PIO A
	PIO_A_CMD = 0xCF; // 0b11001111 Mode 3
	PIO_A_CMD = 0x00; // Output
}
#endif /* Z80_KIO */

void main(void) {

	circbuff_init(&cb_sioa_rx);
	circbuff_init(&cb_sioa_tx);

#if Z80_KIO
	init_kio();
#else
	init_z84c015();
#endif

	init_int_vectors();

	__asm__ ( "ei" );

	LCD_Init();
	LCD_SetCursor(LCD_LINE1);

	const unsigned char *text = "\n\rHello Zilog Z80";

#if 0
	int textlen = 15;
	for (int i = 0; i < textlen; i++) {
		sio_putc(text[i]);
		LCD_Data(text[i]);
	}
#else
    for (const unsigned char *p = text; *p; p++) {
		sio_putc(*p);
		LCD_Data(*p);
	}
#endif
	sio_putc(13); sio_putc(10);

	unsigned char leds = 0xAA;
	const char* hexDigits = "0123456789ABCDEF";

	while (1) {
		int value;
		if ((value = sio_getc()) != EOF) {
			sio_putc(value);
		}

		LCD_SetCursor(LCD_LINE2);
		char test[20];

		__asm__ ( "di" ); /* atomic read */
		unsigned long secs = seconds;
		__asm__ ( "ei" );
		unsigned char hours = secs / 3600;
		secs %= 3600;
		unsigned char mins = secs / 60;
		secs %= 60;

		sprintf(test, "%02u:%02u:%02u", hours, mins, secs);
		for (int i = 0; i < strlen(test); i++)
			LCD_Data(test[i]);

		switch (value) {
		case '0':
			/* Programm unterbrechen, falls Breakpoint gesetzt */
			breakpoint_func(1);
			break;
		case '1':
			/* Text ausgeben, der deutlich länger als der Sendepuffer ist */
			sio_putc('\n'); sio_putc('\r');
			const unsigned char *longtext =
				"0        1         2         3\n\r"
				"123456789012345678901234567890\n\r"
				"+----------------------------+\n\r"
				"| Hello TOSHIBA TMPZ84C015B  |\n\r"
				"+----------------------------+\n\r";
			for (const unsigned char *p = longtext; *p; p++) {
				sio_putc(*p);
			}
			break;
			/* aktuelle Laufzeit ausgeben */
		case '2':
			sio_putc('\n'); sio_putc('\r');
			for (unsigned char *p = test; *p; p++) {
				sio_putc(*p);
			}
			sio_putc('\n'); sio_putc('\r');
			break;
		}

		/*
		sendChar(hexDigits[(counter >> 4) & 0x0F]);
		sendChar(hexDigits[counter & 0x0F]);
		sendChar('-');
		sendChar(hexDigits[(seconds >> 4) & 0x0F]);
		sendChar(hexDigits[seconds & 0x0F]);
		sendChar('-');
		sendChar(13); sendChar(10);
		*/
		/*
		PIA_C_DATA = leds;
		leds = ~leds;
		*/
		// delay(1000);
	}

	/* ------------------------------
	unsigned char ch = 0;
	while (ch != 13) {
		ch = readChar();
		sendChar(ch);
		LCD_Data(ch);
	}
	sendChar(13); sendChar(10);
	--------------------------------*/

	/* ------------------------------
	// LED Lauflicht
	unsigned char leds = 1;
	while (1) {
		for (int i = 0; i < 7; i++) {
			PIA_C_DATA = leds;
			leds = leds << 1;
			wait();
		}
		for (int i = 0; i < 7; i++) {
			PIA_C_DATA = leds;
			leds = leds >> 1;
			wait();
		}
	}
	---------------------------------*/
}
